#!/usr/bin/env python
###############################################################################
# (c) Copyright 2018 CERN                                                     #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
'''
Code to bootstrap the initial environment.
'''


def collect_roots(root):
    '''
    Return an iterator of siteroot directories required by the requested root
    dir.

    Chaining of root directories is recorded in {root}/etc/chaining_infos.json
    as a list of required roots.
    '''
    from os.path import join, isdir
    from json import load
    from itertools import chain, imap

    # only take into account existing directories
    if isdir(root):
        yield root  # the requested entry is always the first of the list
        try:
            # get the list of dependencies
            deps = load(open(join(root, 'etc', 'chaining_infos.json')))
            # recurse in each of them, in order
            for r in chain.from_iterable(imap(collect_roots, deps)):
                yield r
        except (IOError, ValueError):
            pass


def search_path(roots):
    '''
    Generate a list of entries for the search path from a list of siteroot
    dirs.
    '''
    from os.path import join
    suffixes = [
        'lhcb',
        join('lcg', 'releases'),
        # backward compatibility
        join('lcg', 'app', 'releases'),
        join('lcg', 'external'),
        # hint to look up special projects
        'contrib',
        # allow siteroot specific modules
        'cmake'
    ]

    yield '${LHCBRELEASES}'
    yield '${LCG_RELEASES}'

    # see if we have LbDevTools
    try:
        from LbDevTools import DATA_DIR
        yield join(DATA_DIR, 'cmake')
    except ImportError:
        pass

    for root in roots:
        for suff in suffixes:
            yield join(root, suff)


def bin_path(roots, host_os=None, host_flavour=None):
    '''
    Generate the list of directories to be prepended to the PATH variable.
    '''
    from os.path import join

    for root in roots:
        bindir = join(root, 'bin')
        if host_os:
            yield join(bindir, host_os)
        if host_flavour:
            yield join(bindir, host_flavour)
        yield bindir


def main():
    import os
    import sys
    from os.path import join
    from LbEnv import __version__

    from argparse import ArgumentParser
    parser = ArgumentParser()
    parser.add_argument(
        '--version',
        action='version',
        version='%(prog)s {}'.format(__version__))
    parser.add_argument(
        '-c', '--platform', help='override the default platform detection')
    parser.add_argument(
        '-r', '--siteroot', help='path to the installation root')
    parser.add_argument(
        "--sh",
        action="store_const",
        const="sh",
        dest="shell",
        help="print the changes to the environment as shell "
        "commands for 'sh'-derived shells")
    parser.add_argument(
        "--csh",
        action="store_const",
        const="csh",
        dest="shell",
        help="print the changes to the environment as shell "
        "commands for 'csh'-derived shells")
    parser.add_argument(
        "--py",
        action="store_const",
        const="py",
        dest="shell",
        help="print the changes to the environment as a Python "
        "dictionary")

    if (set(['-h', '--help']).intersection(sys.argv)
            and set(['--sh', '--csh', '--py']).intersection(sys.argv)):
        # if we are in shell script mode, and want the help message, it should
        # be printed on stderr
        sys.stdout = sys.stderr

    args = parser.parse_args()

    if os.environ.get('LBENV_SOURCED'):
        # LbEnv already called, do not run again
        return

    import LbEnv
    import LbPlatformUtils
    from xenv.Control import Environment
    control = Environment()
    control.presetFromSystem()
    original_env = dict(os.environ)

    # Make sure this script is sourced exactly once
    control.set('LBENV_SOURCED', '1')

    # make sure we can find data files
    if args.platform is None:
        args.platform = LbEnv.defaultPlatform()

    # this is something like "Linux-x86_64"
    host_flavour = '{0}-{4}'.format(*os.uname())
    host_os = LbPlatformUtils.host_os()

    control.set('BINARY_TAG', args.platform)
    control.set('LCG_hostos', host_os)

    if args.siteroot:
        mysiteroot = args.siteroot
    elif 'MYSITEROOT' in os.environ:
        mysiteroot = os.environ['MYSITEROOT']
    elif 'VO_LHCB_SW_DIR' in os.environ:
        mysiteroot = join(os.environ['VO_LHCB_SW_DIR'], 'lib')
    elif 'VIRTUAL_ENV' in os.environ:
        mysiteroot = os.environ['VIRTUAL_ENV']
        # the venv is usually in #MYSITEROOT/var/..., so we try to deduce it
        head, tail = mysiteroot, 'none'
        while tail and tail != 'var':
            head, tail = os.path.split(head)
        if tail:
            # 'var' was found in the path, so the parent directory is the root
            mysiteroot = head
    else:
        sys.stderr.write('error: not valid siteroot provided\n')
        sys.exit(1)

    control.set('MYSITEROOT', mysiteroot)

    if 'HOME' in os.environ and os.path.exists(
            join(os.environ['HOME'], 'cmtuser')):
        control.set('User_release_area', join(os.environ['HOME'], 'cmtuser'))

    control.set('LHCBRELEASES',
                os.environ.get('LHCB_LOC', join(mysiteroot, 'lhcb')))
    control.set(
        'LCG_RELEASES',
        os.environ.get('LCG_RELEASES', join(mysiteroot, 'lcg', 'releases')))

    roots = list(collect_roots(mysiteroot))

    control.prepend('CMAKE_PREFIX_PATH', ':'.join(search_path(roots)))

    # Add to the paths the utilities hosted in the siteroots
    control.prepend(
        'PATH', ':'.join(
            bin_path(roots, host_os=host_os, host_flavour=host_flavour)))

    if 'VIRTUAL_ENV' in os.environ:
        control.default('XDG_DATA_DIRS', '/usr/local/share:/usr/share')
        control.prepend('XDG_DATA_DIRS',
                        join(os.environ['VIRTUAL_ENV'], 'share'))
        control.prepend('XDG_DATA_DIRS', join(mysiteroot, 'share'))

    # FIXME this chunk is copied from xenv.Script (we need refactoring)
    # print only modified variables
    env = control.vars()

    # generate and print the banner _before_ removing unchanged variables
    if sys.stderr.isatty():
        import LbEnv.Banner
        sys.stderr.write(LbEnv.Banner.generate(env=env))
        sys.stderr.write('\n')

    # only set unmodified variables
    env = dict((name, value) for name, value in env.items()
               if original_env.get(name) != value)

    if args.shell == 'py':
        from pprint import pprint
        pprint(env)
    else:
        template = {
            'sh': "export {0}='{1}'",
            'csh': "setenv {0} '{1}';"
        }.get(args.shell, "{0}={1}")
        print '\n'.join(
            template.format(name, value)
            for name, value in sorted(env.items()))


if __name__ == '__main__':  # pragma no cover
    main()



###############################################################################
# (c) Copyright 2018 CERN                                                     #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################


def test_chained_roots_list():
    from LbEnv.Utils.Temporary import TempDir
    from os import makedirs
    from os.path import join
    from json import dump
    from LbEnv.Bootstrap import collect_roots

    with TempDir() as tmpdir:
        deps = {
            0: None,  # 0 does not contain the json
            1: [0],  # simple
            2: [1],  # 2 levels
            3: [2],  # 3 levels
            4: [],  # empty JSON
            5: [4],  # simple anding on empty JSON
            6: [2, 5],  # multi deps (does it make sense?)
            7: [8],  # chain to non-existend directory
        }
        rootdir = lambda n: join(tmpdir, 'root{}'.format(n))

        for n in deps:
            root = rootdir(n)
            etc = join(root, 'etc')
            makedirs(etc)
            if deps[n] is not None:
                with open(join(etc, 'chaining_infos.json'), 'w') as f:
                    dump([rootdir(n) for n in deps[n]], f)

        assert list(collect_roots(rootdir(0))) == map(rootdir, [0])
        assert list(collect_roots(rootdir(1))) == map(rootdir, [1, 0])
        assert list(collect_roots(rootdir(2))) == map(rootdir, [2, 1, 0])
        assert list(collect_roots(rootdir(3))) == map(rootdir, [3, 2, 1, 0])
        assert list(collect_roots(rootdir(4))) == map(rootdir, [4])
        assert list(collect_roots(rootdir(5))) == map(rootdir, [5, 4])
        assert list(collect_roots(rootdir(6))) == map(rootdir,
                                                      [6, 2, 1, 0, 5, 4])
        assert list(collect_roots(rootdir(7))) == map(rootdir, [7])
        assert list(collect_roots(rootdir(8))) == map(rootdir, [])


def test_search_path():
    # make sure we cannot use LbDevTools
    import sys
    try:
        import LbDevTools
        old = sys.modules['LbDevTools']
        sys.modules['LbDevTools'] = sys.modules['LbEnv']
    except ImportError:
        old = None

    try:
        from os.path import join
        from LbEnv.Bootstrap import search_path
        from LbEnv.Utils.Temporary import TempDir

        base = '/path/to/base'
        derived = '/path/to/derived'

        suffixes = [
            'lhcb', 'lcg/releases', 'lcg/app/releases', 'lcg/external',
            'contrib', 'cmake'
        ]

        result = list(search_path([]))
        expected = ['${LHCBRELEASES}', '${LCG_RELEASES}']
        assert result == expected

        result = list(search_path([base]))
        expected = ['${LHCBRELEASES}', '${LCG_RELEASES}'] + \
            [join(base, suff) for suff in suffixes]
        assert result == expected

        result = list(search_path([derived, base]))
        expected = ['${LHCBRELEASES}', '${LCG_RELEASES}'] + \
            [join(derived, suff) for suff in suffixes] + \
            [join(base, suff) for suff in suffixes]
        assert result == expected

        with TempDir() as tmpdir:
            sys.path.insert(0, tmpdir)
            try:
                if 'LbDevTools' in sys.modules:
                    del sys.modules['LbDevTools']

                with open(join(tmpdir, 'LbDevTools.py'), 'w') as f:
                    f.write('DATA_DIR = "/path/to/devtools"\n')

                result = list(search_path([base]))
                expected = ['${LHCBRELEASES}', '${LCG_RELEASES}',
                            '/path/to/devtools/cmake'] + \
                    [join(base, suff) for suff in suffixes]
                assert result == expected

            finally:
                sys.path.pop(0)

                if 'LbDevTools' in sys.modules:
                    del sys.modules['LbDevTools']

    finally:
        if old:
            sys.modules['LbDevTools'] = old


def test_bin_path():
    from os.path import join
    from LbEnv.Bootstrap import bin_path

    base = '/path/to/base'
    derived = '/path/to/derived'

    result = list(bin_path([], 'host-os', 'host-flavour'))
    expected = []
    assert result == expected

    result = list(bin_path([base]))
    expected = [join(base, 'bin')]
    assert result == expected

    result = list(bin_path([base], host_os='host-os'))
    expected = [join(base, 'bin', 'host-os'), join(base, 'bin')]
    assert result == expected

    result = list(bin_path([base], host_flavour='host-flavour'))
    expected = [join(base, 'bin', 'host-flavour'), join(base, 'bin')]
    assert result == expected

    result = list(
        bin_path([base], host_os='host-os', host_flavour='host-flavour'))
    expected = [
        join(base, 'bin', 'host-os'),
        join(base, 'bin', 'host-flavour'),
        join(base, 'bin')
    ]
    assert result == expected

    result = list(
        bin_path(
            [derived, base], host_os='host-os', host_flavour='host-flavour'))
    expected = [
        join(derived, 'bin', 'host-os'),
        join(derived, 'bin', 'host-flavour'),
        join(derived, 'bin'),
        join(base, 'bin', 'host-os'),
        join(base, 'bin', 'host-flavour'),
        join(base, 'bin')
    ]
    assert result == expected



###############################################################################
# (c) Copyright 2018 CERN                                                     #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
'''
@author: Marco Clemencic <marco.clemencic@cern.ch>
'''

import os
import re
import logging

# FIXME: when we drop Python 2.4, this should become 'from . import path'
from LbEnv.ProjectEnv import path, Error
from LbEnv.ProjectEnv.version import (DEFAULT_VERSION, versionKey as _vkey,
                                      LCGInfoName)

log = logging.getLogger(__name__)

_PLATFORM_ID_RE = re.compile(
    r'((x86_64|i686)-[a-z]+[0-9]+-[a-z]+[0-9]+-[a-z0-9]+)|'
    r'([a-z]+[0-9]+_[a-z]+[0-9]+_[a-z]+[0-9]+(_dbg)?)')

# use uppercase names for case insensitive match
EXTERNAL_PROJECTS = ('ROOT', )


def platform_sort_key(platform):
    '''
    Key function to sort platforms.

    The preferred platform has highest score.
    '''
    if '-' not in platform:
        os_id, arch, comp = platform.split('_')
        arch = {'ia32': 'i686', 'amd64': 'x86_64'}.get(arch, arch)
        opt = 'opt'
    else:
        arch, os_id, comp, opt = platform.split('-')
    os_id = tuple(int(i) for i in os_id if i.isdigit())
    comp = tuple(int(i) for i in comp if i.isdigit())
    return ('0' if opt == 'do0' else opt, arch, comp, os_id)


def isPlatformId(s):
    '''
    Return True if the string looks like a platform id.

    >>> isPlatformId('x86_64-centos7-gcc64-opt')
    True
    >>> isPlatformId('slc4_ia32_gcc34')
    True
    >>> isPlatformId('include')
    False
    '''
    return bool(_PLATFORM_ID_RE.match(s))


def versionKey(x):
    return _vkey(x[0])


class NotFoundError(Error):
    '''
    Generic error for configuration elements that are not found.
    '''

    def __str__(self):
        return 'cannot find {0}'.format(self.args[0])


class MissingManifestError(NotFoundError):
    '''
    The manifest.xml for a project was not found.
    '''


class MissingProjectError(NotFoundError):
    '''
    A project was not found.
    '''

    def __init__(self, *args):
        super(MissingProjectError, self).__init__(*args)
        self.name, self.version, self.platform, self.path = args

    def __str__(self):
        return 'cannot find project {0} {1} for {2} in {3}'.format(*self.args)


class MissingDataPackageError(NotFoundError):
    '''
    A data package was not found.
    '''

    def __init__(self, *args):
        super(MissingDataPackageError, self).__init__(*args)
        self.name, self.version, self.path = args

    def __str__(self):
        return 'cannot find data package {0} {1} in {2}'.format(*self.args)


class InvalidNightlySlotError(NotFoundError):
    '''
    A nightly slot build was not found.
    '''

    def __init__(self, *args):
        super(InvalidNightlySlotError, self).__init__(*args)
        self.slot, self.build_id, self.path = args

    def __str__(self):
        return 'cannot find nightly slot build {0}/{1}{2}.'.format(
            self.args[0], self.args[1], ' in {}'.format(self.args[2])
            if self.args[2] else '')


def findFile(name, search_path=None):
    '''
    Look for a file in the search path.
    '''
    log.debug('looking for file %r', name)
    from LbEnv import which
    fn = which(name, search_path or path)
    if fn:
        log.debug('found %r', fn)
    else:
        raise NotFoundError(name)
    return fn


def findProject(name, version, platform, allow_empty_version=False):
    '''
    Find a Gaudi-based project in the directories specified in the 'path'
    variable.

    @param name: name of the project (case sensitive for local projects)
    @param version: version of the project
    @param platform: binary platform id
    @param allow_empty_version: if True, we allow also the plain project name
                                (without version)

    @return path to the project binary directory

    If name is None, version should be the path to the top directory of the
    project.
    '''
    log.debug('findProject(name=%r, version=%r, platform=%r, '
              'allow_empty_version=%r)', name, version, platform,
              allow_empty_version)

    if name:
        # standard project suffixes
        suffixes = [
            '{0}_{1}'.format(name, version),
            os.path.join(name.upper(), '{0}_{1}'.format(name.upper(),
                                                        version)),
            os.path.join(name.upper(), version)
        ]
        # special case: for the 'latest' version we allow the plain name
        if allow_empty_version:
            suffixes.insert(0, name)
    else:
        # path to project used, no need for suffixes
        suffixes = [os.curdir]

    # if project name is None, version is the path to the top level directory
    # of the project
    for d in [
            os.path.normpath(os.path.join(b, s, bindir))
            for b in (path if name else [version]) for s in suffixes
            for bindir in (os.path.join('InstallArea', platform), os.curdir)
    ]:
        log.debug('check %s', d)
        if os.path.exists(d):
            log.debug('OK')
            return d
    else:
        raise MissingProjectError(name, version, platform, path)


def findNightlyDir(slot, build_id, nightly_bases):
    '''
    Return the directory of the requested build of a nightly slot, looking in
    the directories listed in nightly_bases.

    If not found raise InvalidNightlySlotError
    '''
    # FIXME: we cannot use logging here because this function is called too
    # early in the script

    # log.debug('looking for slot %s %s', slot, build_id)
    for slot_dir in [
            os.path.join(nightly_base, slot_id, build_id)
            for slot_id in (slot, 'lhcb-' + slot)
            for nightly_base in nightly_bases
    ]:
        if os.path.isdir(slot_dir):
            # log.debug('found %s', slot_dir)
            return slot_dir
    # log.warning('not found')
    raise InvalidNightlySlotError(slot, build_id, nightly_bases)


def listVersions(name, platform):
    '''
    Find all instances of a Gaudi-based project in the directories specified in
    the 'path' variable and return the list of versions found.

    @param name: name of the project (case sensitive for local projects)
    @param platform: binary platform id

    @return generator of pairs (version, fullpath)
    '''
    # for special external projects case we delegate to another function
    if name.upper() in EXTERNAL_PROJECTS:
        for entry in listExtVersions(name, platform):
            yield (entry[0], entry[2])
        return
    # FIXME: when we drop Python 2.4, this should become
    # 'from .version import isValidVersion'
    from LbEnv.ProjectEnv.version import isValidVersion
    log.debug('listVersions(name=%r, platform=%r)', name, platform)

    name_u = name.upper()
    prefix = name + '_'
    prefix_u = name_u + '_'
    prefixlen = len(prefix)

    signature = os.path.join('InstallArea', platform, os.curdir)
    if name == 'LCG':
        signature = LCGInfoName(platform)
    log.debug('looking for %s in the search path', signature)

    def matches(path):
        # log.debug('testing %s', path)
        return (os.path.exists(os.path.join(path, signature))
                or platform == 'any')

    found_versions = set()
    for p in [dirname for dirname in path if os.path.isdir(dirname)]:
        files = set(os.listdir(p))
        # the plain project name is taken into account as 'default' version
        if DEFAULT_VERSION not in found_versions and name in files:
            fullpath = os.path.join(p, name)
            if matches(fullpath):
                found_versions.add(DEFAULT_VERSION)
                yield (DEFAULT_VERSION, fullpath)

        # versions like 'Project_vXrY'
        for entry in sorted(
            [(filename[prefixlen:], os.path.join(p, filename))
             for filename in files
             if (filename.startswith(prefix)
                 and isValidVersion(name, filename[prefixlen:]))],
                reverse=True,
                key=versionKey):
            version, fullpath = entry
            if (version not in found_versions and matches(fullpath)):
                found_versions.add(version)
                yield entry

        # versions like PROJECT/PROJECT_vXrY
        project_dir = os.path.join(p, name_u)
        if os.path.isdir(project_dir):
            for entry in sorted(
                [(filename[prefixlen:], os.path.join(project_dir, filename))
                 for filename in os.listdir(project_dir)
                 if (filename.startswith(prefix_u)
                     and isValidVersion(name, filename[prefixlen:]))],
                    reverse=True,
                    key=versionKey):
                version, fullpath = entry
                if (version not in found_versions and matches(fullpath)):
                    found_versions.add(version)
                    yield entry


def listPlatforms(name, version, allow_empty_version=False, quiet=False):
    '''
    Find a version of a Gaudi-based project in the directories specified in
    the 'path' variable and return the list of platforms available.

    @param name: name of the project (case sensitive for local projects)
    @param version: version string
    @allow_empty_version: search also in directories without the version
    @quiet: if true, do not print a warning if the project is not found

    @return list of platform strings

    If name is None, version should be the path to the top directory of the
    project.
    '''
    from os.path import isdir, join, normpath, exists
    # for special external projects case we delegate to another function
    if name.upper() in EXTERNAL_PROJECTS:
        return listExtPlatforms(name, version)
    log.debug('listPlatforms(name=%r, version=%r, allow_empty_version=%r)',
              name, version, allow_empty_version)

    if name:
        # standard project suffixes
        suffixes = [
            '{0}_{1}'.format(name, version),
            join(name.upper(), '{0}_{1}'.format(name.upper(), version)),
            join(name.upper(), version)
        ]
        # special case: for the 'latest' version we allow the plain name
        if allow_empty_version:
            suffixes.insert(0, name)
    else:
        # path to project used, no need for suffixes
        suffixes = [os.curdir]

    platforms = set()
    # if project name is None, version is the path to the top level directory
    # of the project
    for d in [
            normpath(join(b, s)) for b in (path if name else [version])
            for s in suffixes
    ]:
        log.debug('check %s', d)
        inst_area = join(d, 'InstallArea')
        if isdir(inst_area):
            for platform in os.listdir(inst_area):
                p_dir = join(inst_area, platform)
                if (isdir(p_dir) and isPlatformId(platform)
                        or exists(join(p_dir, 'manifest.xml'))):
                    platforms.add(platform)
        elif exists(join(d, 'manifest.xml')):
            platforms.add('<no-platform>')

    if not platforms:
        (log.debug if quiet else log.warning)('could not find %s/%s in %r',
                                              name, version, path)
    return sorted(platforms, key=platform_sort_key, reverse=True)


def findDataPackage(name, version):
    '''
    Find a data package in the directories specified in the 'path' variable,
    using, optionally, the standard suffixes 'DBASE' and 'PARAM'.
    If version is a pattern, the latest version matching the pattern is
    returned.

    @param name: name of the package with "hat" (case sensitive)
    @param version: glob pattern to filter the version

    @return: the path to the data package
    '''
    from fnmatch import fnmatch

    suffixes = ['', 'DBASE', 'PARAM']
    versions = []
    for p in [os.path.join(b, s, name) for b in path for s in suffixes]:
        if os.path.exists(p):
            lst = os.listdir(p)
            if version in lst:
                # stop searching if we find an exact match
                return os.path.join(p, version)
            versions.extend([(v, p) for v in lst if fnmatch(v, version)])
    if not versions:
        raise MissingDataPackageError(name, version, path)
    # sort the versions found
    versions.sort(key=versionKey, reverse=True)
    v, p = versions[0]
    return os.path.join(p, v)


def parseManifest(manifest):
    '''
    Extract the list of required projects and data packages from a manifest.xml
    file.

    @param manifest: path to the manifest file
    @return: tuple with ([projects...], [data_packages...]) as (name, version)
             pairs
    '''
    from xml.dom.minidom import parse
    m = parse(manifest)

    def _iter(parent, child):
        '''
        Iterate over the tags in <parent><child/><child/></parent>.
        '''
        for pl in m.getElementsByTagName(parent):
            for c in pl.getElementsByTagName(child):
                yield c

    # extract the list of used (project, version) from the manifest
    used_projects = [(p.attributes['name'].value,
                      p.attributes['version'].value)
                     for p in _iter('used_projects', 'project')]
    # extract the list of data packages
    data_packages = [(p.attributes['name'].value,
                      p.attributes['version'].value)
                     for p in _iter('used_data_pkgs', 'package')]
    return (used_projects, data_packages)


def getEnvXmlPath(project, version, platform, allow_empty_version=False):
    '''
    Return the list of directories to be added to the Env XML search path for
    a given project.
    '''
    pdir = findProject(
        project, version, platform, allow_empty_version=allow_empty_version)
    search_path = [pdir]
    # manifests to parse
    manifests = [os.path.join(pdir, 'manifest.xml')]
    while manifests:
        manifest = manifests.pop(0)
        if not os.path.exists(manifest):
            raise MissingManifestError(manifest)
        projects, packages = parseManifest(manifest)
        # add the data package directories
        search_path.extend([findDataPackage(p, v) for p, v in packages])
        # add the project directories ...
        pdirs = [findProject(p, v, platform) for p, v in projects]
        search_path.extend(pdirs)
        # ... and their manifests to the list of manifests to parse
        manifests.extend(
            [os.path.join(pdir, 'manifest.xml') for pdir in pdirs])

    def _unique(iterable):
        returned = set()
        for i in iterable:
            if i not in returned:
                returned.add(i)
                yield i

    return list(_unique(search_path))


def findLCG(version, platform):
    '''
    Return the path to the requested LCG version, found in the search path.
    '''
    for p in [
            os.path.join(b, s, n) for b in path
            for s in ('', 'LCG_%s' % version)
            for n in ('LCG_%s_%s.txt' % (version, platform),
                      LCGInfoName(platform))
    ]:
        if os.path.exists(p):
            return os.path.dirname(p)
    for p in [
            os.path.join(b, 'LCGCMT', s) for b in path
            for s in ('LCGCMT_%s' % version, 'LCGCMT-%s' % version)
    ]:
        if os.path.exists(p):
            return p


def getHepToolsInfo(manifest):
    '''
    Extract the hep tools version and platform from a manifest file.
    '''
    from xml.dom.minidom import parse

    log.debug('extracting heptools version from %s', manifest)
    try:
        m = parse(manifest)
        heptools = m.getElementsByTagName('heptools')[0]
        version = (heptools.getElementsByTagName('version')[0]
                   .firstChild.nodeValue)
        platform = (heptools.getElementsByTagName('binary_tag')[0]
                    .firstChild.nodeValue)
        return version, platform
    except (IndexError, AttributeError), exc:
        # cannot extract heptools version and platform
        raise NotFoundError('heptools info: %s' % exc)


def getLCGRelocation(manifest):
    from os.path import dirname, basename, join, pardir

    try:
        version, platform = getHepToolsInfo(manifest)
    except NotFoundError, exc:
        log.debug(str(exc))
        return {}

    log.debug('looking for LCG %s %s', version, platform)
    lcg_path = findLCG(version, platform)
    if not lcg_path:
        log.debug('cannot find LCG path')
        return {}
    # FIXME: xenv has got problems with unicode filenames
    lcg_path = str(lcg_path)
    log.debug('found LCG at %s', lcg_path)

    if basename(dirname(lcg_path)) == 'LCGCMT':
        # old style
        lcg_root = dirname(dirname(lcg_path))
        return {
            'LCG_releases':
            lcg_root,
            'LCG_external': (lcg_root
                             if lcg_root.endswith('external') else join(
                                 lcg_root, pardir, pardir, 'external'))
        }
    else:
        # new style
        return {
            'LCG_releases':
            lcg_path,
            'LCG_external':
            lcg_path,
            'LCG_releases_base': (dirname(lcg_path) if lcg_path.endswith(
                'LCG_%s' % version) else lcg_path)
        }


def getProjectNameVersion(manifest):
    '''
    Get project name and version from a manifest.xml.
    '''
    from xml.dom.minidom import parse

    log.debug('extracting project name from %s', manifest)
    name, version = None, None
    try:
        m = parse(manifest)
        project = m.getElementsByTagName('project')[0]
        name = project.getAttribute('name')
        version = project.getAttribute('version')
    except (IndexError, AttributeError), exc:
        log.debug('cannot extract project name or version: %s', exc)
    return name, version


def listExtVersions(ext, platform):
    '''
    @return generator of tuples (ext_version, LCG_version, LCG_path)
    '''
    ext = ext.upper()  # case insensitive check
    found_versions = set()
    for LCG_version, LCG_path in listVersions('LCG', platform):
        for l in open(os.path.join(LCG_path, LCGInfoName(platform))):
            l = l.split(';')
            if l[0].strip().upper() == ext:
                v = l[2].strip()
                if v not in found_versions:
                    found_versions.add(v)
                    yield (l[2].strip(), LCG_version, LCG_path)
                break  # go to next LCG version


def listExtPlatforms(ext, version):
    '''
    @return list of available platforms for a version of an extrenal project
    '''
    import glob
    log.debug('listExtPlatforms(ext=%r, version=%r)', ext, version)

    ext = ext.upper()  # case insensitive check

    info_glob = LCGInfoName('*')
    # transform the glob in a capturing regexp
    platform_exp = re.compile(re.escape(info_glob).replace('\\*', '(.*)'))

    platforms = set()
    for p in [
            os.path.join(b, s, info_glob) for b in path for s in ('', 'LCG_*')
    ]:
        for f in glob.glob(p):
            # it must match if the glob matched
            platform = platform_exp.search(f).group(1)
            # to be sure, check if the external is in the list with the right
            # version
            for l in (l.split(';') for l in open(f)):
                if l[0].strip().upper() == ext and l[2].strip() == version:
                    platforms.add(platform)
                    break
    return sorted(platforms, key=platform_sort_key, reverse=True)


def findLCGForExt(ext, version, platform):
    '''
    Return the (highest) version of LCG containing the required version of
    an external project.
    '''
    for ext_vers, LCG_vers, _path in listExtVersions(ext, platform):
        if ext_vers == version:
            return LCG_vers
    raise MissingProjectError(ext, version, platform, path)


PREFERRED_PLATFORM = (os.environ.get('BINARY_TAG')
                      or os.environ.get('CMTCONFIG') or None)


def _findManifest(path, platform_hint=PREFERRED_PLATFORM):
    '''
    Find a manifest file of the project rooted at 'path'.
    '''
    from os.path import join, exists, isdir
    # try the InstallArea level for the current platform
    inst = join(path, 'InstallArea')

    def candidates():
        if platform_hint:
            yield join(inst, platform_hint, 'manifest.xml')
        if isdir(inst):
            for p in os.listdir(inst):
                yield join(inst, p, 'manifest.xml')
        yield join(path, 'manifest.xml')

    for c in candidates():
        if exists(c):
            return c
    return None


def walkProjectDeps(project, version, platform='any', ignore_missing=False):
    '''
    Return a tuple ((project, version), rootdir, dependencies) for the
    requested project and all its dependencies.

    The dependencies are returned as a list of (project, version) and the list
    can be altered to control which dependency to follow and which not.

    Note that when walking through the dependencies a project may appear more
    than once (for diamond dependency graphs).
    '''
    if project.lower() in ('lcg', 'heptools', 'lcgcmt'):
        return
    try:
        root = findProject(project, version, platform)
    except MissingProjectError as err:
        if ignore_missing:
            log.warning(str(err))
            return
        else:
            raise
    manifest = _findManifest(root)
    if manifest:
        deps = [(p, v) for p, v in parseManifest(manifest)[0]]
        try:
            deps.append(('lcg', getHepToolsInfo(manifest)[0]))
        except NotFoundError:
            pass
    else:
        deps = []
    yield (project, version), root, deps
    for project, version in deps:
        for x in walkProjectDeps(project, version, platform, ignore_missing):
            yield x


def getLHCbGrid(platform):
    '''
    Find the best LHCbGrid version and platform for the requested platform.
    '''
    from LbEnv.ProjectEnv.version import expandVersionAlias
    for version in [DEFAULT_VERSION, 'latest']:
        version = expandVersionAlias('LHCbGrid', 'latest', 'any')
        platforms = listPlatforms('LHCbGrid', version, quiet=True)
        if platforms:
            break
    else:
        raise NotFoundError('LHCbGrid')
    if platform not in platforms:
        log.warning('platform %s not available for LHCbGrid', platform)
        os_id = platform.split('-')[1]
        for platform in platforms:
            if platform.split('-')[1] == os_id:
                log.warning('using %s instead', platform)
                break
    return version, platform



###############################################################################
# (c) Copyright 2018 CERN                                                     #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
'''
Common functions to add common options to a OptionParser instance.
'''
__author__ = 'Marco Clemencic <marco.clemencic@cern.ch>'

import os
import re

from LbEnv.ProjectEnv import SearchPathEntry, SearchPath
from LbEnv.ProjectEnv.options import LHCbDevPathEntry, NightlyPathEntry


def addSearchPath(parser):
    '''
    Common options used to extend the search path.
    '''
    from optparse import OptionValueError

    def dev_dir_cb(option, opt_str, value, parser):
        if value is None:
            try:
                value = LHCbDevPathEntry()
            except ValueError:
                raise OptionValueError(
                    '--dev used, but LHCBDEV is not defined')
        else:
            value = SearchPathEntry(value)
        parser.values.dev_dirs.append(value)

    parser.add_option(
        '--dev',
        action='callback',
        callback=dev_dir_cb,
        help='prepend $LHCBDEV to the search path. '
        'Note: the directories are searched in the '
        'order specified on the command line.')
    parser.add_option(
        '--dev-dir',
        action='callback',
        metavar='DEVDIR',
        type='string',
        callback=dev_dir_cb,
        help='prepend DEVDIR to the search path. '
        'Note: the directories are searched in the '
        'order specified on the command line.')

    def nightly_base(option, opt_str, value, parser):
        '''
        Callback for the --nightly-base and --nightly-cvmfs options.
        '''
        if parser.values.nightly:
            raise OptionValueError('%s specified after --nightly' % option)

        if not os.path.isdir(value):
            raise OptionValueError('"%s" is not a directory' % value)

        parser.values.nightly_bases.append(value)

    parser.add_option(
        '--nightly-base',
        action='callback',
        type='string',
        callback=nightly_base,
        help='add the specified directory to the nightly builds '
        'search path (must be specified before --nightly)')

    def nightly_option(_option, opt_str, _value, _parser):
        valid_value = re.compile(
            r'^(mon|tue|wed|thu|fri|sat|sun|today|yesterday|latest|'
            '\d{4}-\d\d-\d\d|\d+)$', re.IGNORECASE)
        day = 'Today'

        parser.values.dev = True
        rargs = parser.rargs

        try:
            slot = rargs.pop(0)
        except IndexError:
            raise OptionValueError(
                '%s must be followed by the slot of the '
                'nightlies and optionally by the build id' % opt_str)

        if '/' in slot:
            slot, day = slot.split('/', 1)
            if valid_value.match(day):
                day = day.capitalize()
        elif rargs:
            match = valid_value.match(rargs[0])
            if match:
                day = rargs.pop(0).capitalize()
                import logging
                logging.warning(
                    'deprecated slot id specification: use "... '
                    '--nightly %s/%s ..." instead', slot, day)
        if day == 'Latest':  # special case
            day = 'latest'

        # Locate the requested slot in the know nightlies directories
        from os import environ
        nightly_bases = parser.values.nightly_bases
        nightly_bases += [
            environ.get('LHCBNIGHTLIES', '/cvmfs/lhcbdev.cern.ch/nightlies'),
            environ.get('LCG_nightlies_area',
                        '/cvmfs/sft-nightlies.cern.ch/lcg/nightlies')
        ]

        from LbEnv.ProjectEnv.lookup import (findNightlyDir,
                                             InvalidNightlySlotError)
        try:
            slot_dir = findNightlyDir(slot, day, nightly_bases)

            nightly_base, slot, day = slot_dir.rsplit(os.sep, 2)
            parser.values.dev_dirs.append(
                NightlyPathEntry(nightly_base, slot, day))
            parser.values.nightly = (slot, day, nightly_base)
        except InvalidNightlySlotError as err:
            # to be able to print a friendly message about local installation
            # of a nightly slot, we cannot exit while parsing the arguments
            parser.values.nightly = err

    parser.add_option(
        '--nightly',
        action='callback',
        metavar='SLOT[/DAY]',
        type='string',
        callback=nightly_option,
        nargs=0,
        help='Add the required slot of the LHCb nightly builds to the list of '
        'DEV dirs. DAY must be a 3 digit abbreviation of the weekday "Today", '
        'an ISO formatted date or an integer, the default is Today. Special '
        'settings of the CMTPROJECTPATH needed for the nightly build slot are '
        'taken into account.')

    parser.add_option(
        '--help-nightly-local',
        action='store_true',
        help='Print instructions on how to install locally and use a nightly '
        'slot build')

    parser.add_option(
        '--user-area',
        action='store',
        help='Use the specified path as User_release_area instead of '
        'the value of the environment variable.')

    parser.add_option(
        '--no-user-area',
        action='store_true',
        help='Ignore the user release area when looking for projects.')

    parser.set_defaults(
        dev_dirs=SearchPath([]),
        user_area=os.environ.get('User_release_area'),
        no_user_area=False,
        nightly=None,
        nightly_bases=[])

    return parser


def addOutputLevel(parser):
    '''
    Add options to get more or less messages.
    '''
    import logging
    parser.add_option(
        '--verbose',
        action='store_const',
        const=logging.INFO,
        dest='log_level',
        help='print more information')
    parser.add_option(
        '--debug',
        action='store_const',
        const=logging.DEBUG,
        dest='log_level',
        help='print debug messages')
    parser.add_option(
        '--quiet',
        action='store_const',
        const=logging.WARNING,
        dest='log_level',
        help='print only warning messages (default)')

    parser.set_defaults(log_level=logging.WARNING)

    return parser


def addPlatform(parser):
    '''
    Add option to specify a platform.
    '''

    parser.add_option(
        '-c', '--platform', help='runtime platform [default: %default]')
    parser.add_option(
        '--force-platform',
        action='store_true',
        help='ignore platform compatibility check')

    parser.set_defaults(platform=None, force_platform=False)

    return parser


def addListing(parser):
    '''
    Add option to request the list of versions.
    '''

    parser.add_option(
        '-l',
        '--list',
        action='store_true',
        help='list the available versions of the requested '
        'project and exit')
    parser.add_option(
        '-L',
        '--list-platforms',
        action='store_true',
        help='list the available platforms for the requested '
        'project/version and exit')

    parser.set_defaults(list=False, list_platforms=False)

    return parser



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2018 CERN                                                     #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
'''
Generate and print the environment banner
'''


def _center(txt, width):
    '''
    Return a string of length `width` with the string `txt` centered

    >>> _center('abc', 14)
    '     abc      '
    '''
    left = ' ' * int((width - len(txt)) / 2)
    right = ' ' * (width - len(txt) - len(left))
    return '%s%s%s' % (left, txt, right)


def generate(width=80, env=None):
    '''
    Return the environment banner string.
    '''
    from os import environ, pathsep
    if env is None:
        env = environ

    lines = []
    lines.append('*' * width)
    lines.append('*%s*' % _center('---- LbEnv ----', width - 2))
    if 'BINARY_TAG' in env:
        lines.append('*%s*' % _center(
            'Using system configuration %s' % env['BINARY_TAG'], width - 2))
    lines.append('*' * width)
    if 'User_release_area' in env:
        lines.append(
            ' --- User_release_area is set to %s' % env['User_release_area'])
    if env.get('CMAKE_PREFIX_PATH'):
        lines.append(' --- CMAKE_PREFIX_PATH is set to:')
        lines.extend('    %s' % item
                     for item in env['CMAKE_PREFIX_PATH'].split(pathsep))
    else:
        lines.append(' --- CMAKE_PREFIX_PATH is NOT set!')
    lines.append('-' * width)
    return '\n'.join(lines)


if __name__ == '__main__':
    print(generate())



###############################################################################
# (c) Copyright 2018 CERN                                                     #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
__author__ = 'Marco Clemencic <marco.clemencic@cern.ch>'

import os
import sys
import shutil
from tempfile import mkdtemp


def setup():
    from os.path import join, dirname, pardir, abspath
    # FIXME compatibility py2-py3
    if sys.version_info >= (3, ):
        os.environ['CMAKE_PREFIX_PATH'] = abspath(
            join(
                dirname(__file__), pardir, pardir, pardir, pardir, pardir,
                'python', 'LbEnv', 'ProjectEnv', 'tests', 'data'))
    else:
        os.environ['CMAKE_PREFIX_PATH'] = join(dirname(__file__), 'data')


latest_gaudi = 'v27r1'


def test_platform_sorting():
    from LbEnv.ProjectEnv.lookup import platform_sort_key
    sorted_platforms = [
        'i686-slc5-gcc43-dbg',
        'i686-slc5-gcc43-opt',
        'slc4_amd64_gcc34',
        'slc4_ia32_gcc34',
        'x86_64-centos7-gcc49-dbg',
        'x86_64-centos7-gcc49-opt',
        'x86_64-centos7-gcc62-dbg',
        'x86_64-centos7-gcc62-do0',
        'x86_64-centos7-gcc62-opt',
        'x86_64-centos7-gcc7-dbg',
        'x86_64-centos7-gcc7-do0',
        'x86_64-centos7-gcc7-opt',
        'x86_64-slc5-gcc43-dbg',
        'x86_64-slc5-gcc43-opt',
        'x86_64-slc5-gcc46-dbg',
        'x86_64-slc5-gcc46-opt',
        'x86_64-slc6-gcc46-dbg',
        'x86_64-slc6-gcc46-opt',
        'x86_64-slc6-gcc48-dbg',
        'x86_64-slc6-gcc48-do0',
        'x86_64-slc6-gcc48-opt',
        'x86_64-slc6-gcc49-dbg',
        'x86_64-slc6-gcc49-do0',
        'x86_64-slc6-gcc49-opt',
        'x86_64-slc6-gcc62-dbg',
        'x86_64-slc6-gcc62-do0',
        'x86_64-slc6-gcc62-opt',
        'x86_64+avx2+fma-centos7-gcc62-dbg',
        'x86_64+avx2+fma-centos7-gcc62-opt',
    ]
    expected_order = [
        'x86_64-slc6-gcc48-do0',
        'x86_64-slc6-gcc49-do0',
        'x86_64-slc6-gcc62-do0',
        'x86_64-centos7-gcc62-do0',
        'x86_64-centos7-gcc7-do0',
        'i686-slc5-gcc43-dbg',
        'x86_64-slc5-gcc43-dbg',
        'x86_64-slc5-gcc46-dbg',
        'x86_64-slc6-gcc46-dbg',
        'x86_64-slc6-gcc48-dbg',
        'x86_64-slc6-gcc49-dbg',
        'x86_64-centos7-gcc49-dbg',
        'x86_64-slc6-gcc62-dbg',
        'x86_64-centos7-gcc62-dbg',
        'x86_64-centos7-gcc7-dbg',
        'x86_64+avx2+fma-centos7-gcc62-dbg',
        'slc4_ia32_gcc34',
        'i686-slc5-gcc43-opt',
        'slc4_amd64_gcc34',
        'x86_64-slc5-gcc43-opt',
        'x86_64-slc5-gcc46-opt',
        'x86_64-slc6-gcc46-opt',
        'x86_64-slc6-gcc48-opt',
        'x86_64-slc6-gcc49-opt',
        'x86_64-centos7-gcc49-opt',
        'x86_64-slc6-gcc62-opt',
        'x86_64-centos7-gcc62-opt',
        'x86_64-centos7-gcc7-opt',
        'x86_64+avx2+fma-centos7-gcc62-opt',
    ]
    sorted_platforms.sort(key=platform_sort_key)
    assert sorted_platforms == expected_order


def prepare_tree(base, tree):
    from os import makedirs
    from os.path import dirname, join, exists
    if hasattr(tree, 'items'):
        tree = tree.items()
    for k, v in tree:
        k = join(base, k)
        if v:
            d = dirname(k)
            if not exists(d):
                makedirs(d)
            f = open(k, 'w')
            f.write(v)
            f.close()
        else:
            if not exists(k):
                makedirs(k)


class TempDir(object):
    def __init__(self, tree=None):
        self.tmpdir = None
        self.tree = tree

    def __enter__(self):
        self.tmpdir = mkdtemp()
        if self.tree:
            prepare_tree(self.tmpdir, self.tree)
        return self.tmpdir

    def __exit__(self, type_, value, traceback):
        if self.tmpdir:
            shutil.rmtree(self.tmpdir)


def test_import():
    import LbEnv.ProjectEnv
    assert LbEnv.ProjectEnv.path


def test_version():
    from LbEnv.ProjectEnv.version import expandVersionAlias, isValidVersion

    assert expandVersionAlias('Gaudi', 'latest',
                              'x86_64-slc6-gcc46-opt') == 'v25r0'

    assert isValidVersion('Gaudi', 'latest')
    assert isValidVersion('Gaudi', 'HEAD')
    assert isValidVersion('Gaudi', latest_gaudi)
    assert isValidVersion('LHCb', 'v32r5p1')
    assert isValidVersion('LHCb', 'v32r4g1')
    assert isValidVersion('LCGCMT', '64d')
    assert not isValidVersion('Gaudi', 'a random string')


# FIXME this is an integration test...
def _test_compatibility():
    from LbEnv.ProjectEnv.compatibility import getOldEnvironment
    env_bk = dict(os.environ)

    try:
        os.environ['CMTCONFIG'] = 'x86_64-slc6-gcc49-opt'
        if 'BINARY_TAG' in os.environ:
            del os.environ['BINARY_TAG']
        env = getOldEnvironment(['Gaudi', latest_gaudi])
        assert 'GAUDISYSROOT' in env
        assert latest_gaudi in env['GAUDISYSROOT']
    finally:
        os.environ.clear()
        os.environ.update(env_bk)

    try:
        os.environ['BINARY_TAG'] = 'x86_64-slc6-gcc49-opt'
        os.environ['CMTCONFIG'] = 'dummy'
        env = getOldEnvironment(['Gaudi', latest_gaudi])
        assert 'GAUDISYSROOT' in env
        assert latest_gaudi in env['GAUDISYSROOT']
    finally:
        os.environ.clear()
        os.environ.update(env_bk)

    try:
        env = getOldEnvironment(['NotAProject', 'v1r0'])
        assert False, 'exception expected'
    except SystemExit, x:
        assert x.code == 1
    finally:
        os.environ.clear()
        os.environ.update(env_bk)


def parse_args(func, args):
    from optparse import OptionParser
    return func(OptionParser(prog='dummy_program')).parse_args(args)


def test_options_addOutputLevel():
    from LbEnv.ProjectEnv.options import addOutputLevel
    import logging

    opts, _ = parse_args(addOutputLevel, [])
    assert opts.log_level == logging.WARNING

    opts, _ = parse_args(addOutputLevel, ['--debug'])
    assert opts.log_level == logging.DEBUG

    opts, _ = parse_args(addOutputLevel, ['--verbose'])
    assert opts.log_level == logging.INFO

    opts, _ = parse_args(addOutputLevel, ['--quiet'])
    assert opts.log_level == logging.WARNING


def test_options_addPlatform():
    from LbEnv.ProjectEnv.options import addPlatform, checkPlatform

    def parse_args(args):
        from optparse import OptionParser
        parser = addPlatform(OptionParser(prog='dummy_program'))
        opts, args = parser.parse_args(args)
        opts.platform = checkPlatform(parser, opts.platform)
        return opts, args

    # set platform, short optino
    opts, _ = parse_args(['-c', 'platform1'])
    assert opts.platform == 'platform1'

    # set platform, long optino
    opts, _ = parse_args(['--platform', 'platform2'])
    assert opts.platform == 'platform2'

    # take CMTCONFIG
    if 'CMTCONFIG' not in os.environ:
        os.environ['CMTCONFIG'] = 'default'
    if 'BINARY_TAG' in os.environ:
        del os.environ['BINARY_TAG']
    opts, _ = parse_args([])
    assert opts.platform == os.environ['CMTCONFIG']

    # prefer BINARY_TAG
    os.environ['BINARY_TAG'] = 'another'
    opts, _ = parse_args([])
    assert opts.platform == os.environ['BINARY_TAG']
    del os.environ['BINARY_TAG']

    # pretend we are on an unsupported platform
    import LbEnv
    bk = LbEnv.defaultPlatform

    def force_known():
        return 'known'

    def force_unknown():
        raise RuntimeError('unknown platform')

    try:
        # no environment, something should be selected
        del os.environ['CMTCONFIG']
        LbEnv.defaultPlatform = force_known
        opts, _ = parse_args([])
        assert opts.platform == 'known'

        LbEnv.defaultPlatform = force_unknown
        opts, _ = parse_args([])
        assert False, 'exception expected'
    except SystemExit:
        pass
    finally:
        LbEnv.defaultPlatform = bk


def test_options_addSearchPath():
    from LbEnv.ProjectEnv.options import addSearchPath
    from LbEnv.ProjectEnv.lookup import InvalidNightlySlotError

    if 'LHCBDEV' not in os.environ:
        os.environ['LHCBDEV'] = '/afs/cern.ch/lhcb/software/DEV'
    opts, _ = parse_args(addSearchPath, ['--dev'])
    assert os.environ['LHCBDEV'] in map(str, opts.dev_dirs)

    del os.environ['LHCBDEV']
    try:
        opts, _ = parse_args(addSearchPath, ['--dev'])
        assert False, 'exception expected'
    except SystemExit:
        pass

    opts, _ = parse_args(addSearchPath, ['--dev-dir', '/some/path'])
    assert '/some/path' in map(str, opts.dev_dirs)

    if 'User_release_area' in os.environ:
        del os.environ['User_release_area']
    opts, _ = parse_args(addSearchPath, [])
    assert opts.user_area is None

    os.environ['User_release_area'] = '/home/myself/cmtuser'
    opts, _ = parse_args(addSearchPath, [])
    assert opts.user_area == os.environ['User_release_area']
    assert not opts.no_user_area

    opts, _ = parse_args(addSearchPath, ['--user-area', '/tmp/myself'])
    assert opts.user_area == '/tmp/myself'

    opts, _ = parse_args(addSearchPath, ['--no-user-area'])
    assert opts.no_user_area

    try:
        opts, _ = parse_args(addSearchPath, ['--nightly'])
        assert False, 'exception expected'
    except SystemExit:
        pass

    import datetime
    days = ('Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat', 'Sun')
    weekday = datetime.date.today().weekday()
    today = 'Today'  # days[weekday]
    yesterday = days[(weekday - 1) % 7]

    with TempDir({
            'lhcb-nightly-slot/%s' % today: None,
            'lhcb-nightly-slot/%s' % yesterday: None
    }) as tmp:
        os.environ['LHCBNIGHTLIES'] = tmp
        opts, _ = parse_args(addSearchPath, ['--nightly', 'lhcb-nightly-slot'])
        assert os.path.join(tmp, 'lhcb-nightly-slot', today) in \
            map(str, opts.dev_dirs)
        assert opts.nightly == ('lhcb-nightly-slot', today, tmp), opts.nightly
        opts, _ = parse_args(addSearchPath,
                             ['--nightly', 'lhcb-nightly-slot', yesterday])
        assert os.path.join(tmp, 'lhcb-nightly-slot', yesterday) in \
            map(str, opts.dev_dirs)
        assert opts.nightly == ('lhcb-nightly-slot', yesterday,
                                tmp), opts.nightly

        opts, _ = parse_args(addSearchPath, ['--nightly', 'dummy'])
        assert isinstance(opts.nightly, InvalidNightlySlotError)

        opts, _ = parse_args(
            addSearchPath,
            ['--nightly', 'lhcb-nightly-slot', days[(weekday + 1) % 7]])
        assert isinstance(opts.nightly, InvalidNightlySlotError)

    with TempDir({
            'lhcb-nightly-slot/%s/confSummary.py' % today:
            '''cmtProjectPathList = ['/extra/dir']\n'''
    }) as tmp:
        os.environ['LHCBNIGHTLIES'] = tmp
        opts, _ = parse_args(addSearchPath, ['--nightly', 'lhcb-nightly-slot'])
        assert '/extra/dir' in map(str, opts.dev_dirs)

    with TempDir({
            'lhcb-nightly-slot/%s/configuration.xml' % today:
            '''<configuration><slot name="lhcb-nightly-slot">
<cmtprojectpath><path value="/extra/xml/dir"/></cmtprojectpath>
</slot></configuration>'''
    }) as tmp:
        os.environ['LHCBNIGHTLIES'] = tmp
        opts, _ = parse_args(addSearchPath, ['--nightly', 'lhcb-nightly-slot'])
        assert '/extra/xml/dir' in map(str, opts.dev_dirs)


def test_lookup():
    from LbEnv.ProjectEnv import path
    bk_path = list(path)

    from LbEnv.ProjectEnv.lookup import (
        findProject, MissingProjectError, findDataPackage,
        MissingDataPackageError, getEnvXmlPath, MissingManifestError)

    if 'BINARY_TAG' in os.environ:
        del os.environ['BINARY_TAG']
    os.environ['CMTCONFIG'] = 'x86_64-slc6-gcc46-opt'

    data = {
        'MYPROJECT/MYPROJECT_v1r0/InstallArea/x86_64-slc6-gcc46-opt/manifest.xml':
        '<?xml version="1.0" encoding="UTF-8"?><manifest></manifest>',
        'DERIVEDPROJECT/DERIVEDPROJECT_v1r3/InstallArea/x86_64-slc6-gcc46-opt/manifest.xml':
        '<?xml version="1.0" encoding="UTF-8"?><manifest>'
        '<used_projects>'
        '<project name="MyProject" version="v1r0" />'
        '</used_projects>'
        '<used_data_pkgs>'
        '<package name="MyPackage" version="*"/>'
        '</used_data_pkgs>'
        '</manifest>',
        'OLDPROJECT/OLDPROJECT_v9r0/InstallArea/x86_64-slc6-gcc46-opt':
        None,
        'DBASE/MyPackage/v2r0':
        None,
        'DBASE/MyPackage/v2r1':
        None,
        'DBASE/MyPackage/v10r0':
        None,
        'User/MyProjectDev_v1r0/InstallArea/x86_64-slc6-gcc46-opt/manifest.xml':
        '<?xml version="1.0" encoding="UTF-8"?><manifest>'
        '<used_projects><project name="MyProject" version="v1r0" /></used_projects>'
        '</manifest>',
        'User/MyAnalysis/InstallArea/x86_64-slc6-gcc46-opt/manifest.xml':
        '<?xml version="1.0" encoding="UTF-8"?><manifest>'
        '<used_projects><project name="MyProject" version="v1r0" /></used_projects>'
        '</manifest>'
    }
    with TempDir(data) as tmp:
        path[:] = [tmp] + bk_path

        proj = findProject('MyProject', 'v1r0', os.environ['CMTCONFIG'])
        assert proj == os.path.join(
            tmp, 'MYPROJECT/MYPROJECT_v1r0/InstallArea/x86_64-slc6-gcc46-opt')

        try:
            findProject('AnotherProject', 'v1r0', os.environ['CMTCONFIG'])
            assert False, 'exception expected'
        except MissingProjectError, x:
            assert str(x)

        # test with user area
        path.insert(0, os.path.join(tmp, 'User'))
        proj = findProject('MyProjectDev', 'v1r0', os.environ['CMTCONFIG'])
        assert proj == os.path.join(
            tmp, 'User/MyProjectDev_v1r0/InstallArea/x86_64-slc6-gcc46-opt')

        proj = findProject(
            'MyAnalysis',
            'latest',
            os.environ['CMTCONFIG'],
            allow_empty_version=True)
        assert proj == os.path.join(
            tmp, 'User/MyAnalysis/InstallArea/x86_64-slc6-gcc46-opt')

        try:
            findProject('MyAnalysis', 'v1r0', os.environ['CMTCONFIG'])
            assert False, 'exception expected'
        except MissingProjectError, x:
            assert str(x)

        pkg = findDataPackage('MyPackage', 'v2r0')
        assert pkg == os.path.join(tmp, 'DBASE/MyPackage/v2r0')

        pkg = findDataPackage('MyPackage', 'v2r*')
        assert pkg == os.path.join(tmp, 'DBASE/MyPackage/v2r1')

        pkg = findDataPackage('MyPackage', 'v*')
        assert pkg == os.path.join(tmp, 'DBASE/MyPackage/v10r0')

        pkg = findDataPackage('MyPackage', '*')
        assert pkg == os.path.join(tmp, 'DBASE/MyPackage/v10r0')

        try:
            findDataPackage('NoPackage', '*')
            assert False, 'exception expected'
        except MissingDataPackageError, x:
            assert str(x)

        xml_path = getEnvXmlPath('DerivedProject', 'v1r3',
                                 os.environ['CMTCONFIG'])
        assert os.path.join(
            tmp, 'MYPROJECT/MYPROJECT_v1r0/InstallArea/x86_64-slc6-gcc46-opt'
        ) in xml_path
        assert os.path.join(
            tmp,
            'DERIVEDPROJECT/DERIVEDPROJECT_v1r3/InstallArea/x86_64-slc6-gcc46-opt'
        ) in xml_path

        try:
            xml_path = getEnvXmlPath('OldProject', 'v9r0',
                                     os.environ['CMTCONFIG'])
            assert False, 'exception expected'
        except MissingManifestError, x:
            assert str(x)

    path[:] = bk_path


def test_listVersions():
    from LbEnv.ProjectEnv import path
    bk_path = list(path)
    from os.path import join

    from LbEnv.ProjectEnv.lookup import listVersions

    if 'BINARY_TAG' in os.environ:
        del os.environ['BINARY_TAG']
    os.environ['CMTCONFIG'] = 'x86_64-slc6-gcc48-opt'

    data = {
        'MYPROJECT/MYPROJECT_v1r0/InstallArea/x86_64-slc6-gcc48-opt/manifest.xml':
        '<?xml version="1.0" encoding="UTF-8"?><manifest></manifest>',
        'MYPROJECT/MYPROJECT_v1r2/InstallArea/x86_64-slc6-gcc48-opt/manifest.xml':
        '<?xml version="1.0" encoding="UTF-8"?><manifest></manifest>'
    }
    with TempDir(data) as tmp:
        path[:] = ['/no/where', tmp, tmp]

        expected = [('v1r2', join(tmp, 'MYPROJECT', 'MYPROJECT_v1r2')),
                    ('v1r0', join(tmp, 'MYPROJECT', 'MYPROJECT_v1r0'))]
        observed = list(listVersions('MyProject', os.environ['CMTCONFIG']))
        print 'expected', expected
        print 'observed', observed
        assert expected == observed

    path[:] = bk_path


def _test_profiling():
    from LbEnv.ProjectEnv import profiling

    from StringIO import StringIO

    try:
        sys.stderr = StringIO()
        profiling.run('sys.exit()')
        assert False, 'exception expected'
    except SystemExit:
        print sys.stderr.getvalue()
    finally:
        sys.stderr = sys.__stderr__

    def dummy():
        raise ImportError()

    import cProfile
    bk1 = cProfile.Profile
    cProfile.Profile = dummy

    try:
        sys.stderr = StringIO()
        profiling.run('sys.exit()')
        assert False, 'exception expected'
    except SystemExit:
        print sys.stderr.getvalue()
    finally:
        sys.stderr = sys.__stderr__

    import profile
    bk2 = profile.Profile
    profile.Profile = dummy
    try:
        sys.stderr = StringIO()
        profiling.run('sys.exit()')
        assert False, 'exception expected'
    except SystemExit:
        print sys.stderr.getvalue()
    finally:
        sys.stderr = sys.__stderr__

    cProfile.Profile = bk1
    profile.Profile = bk2


def test_LBCORE_522():
    '''LBCORE-522: lb-run does not expand some variables when falling back on SetupProject
    https://its.cern.ch/jira/browse/LBCORE-522'''
    from LbEnv.ProjectEnv.compatibility import expandAllVars

    original = {'MYDATA': '${DATA}/subdir', 'DATA': '/main/path'}
    expected = {'MYDATA': '/main/path/subdir', 'DATA': '/main/path'}
    assert expandAllVars(original) == expected

    try:
        expandAllVars({'A': '${B}/x', 'B': '${A}/y'})
        assert False, 'exception expected'
    except ValueError:
        pass


def test_SearchPath():
    from LbEnv.ProjectEnv import (SearchPath, SearchPathEntry,
                                  EnvSearchPathEntry)

    if 'DUMMY_ENV_VAR' in os.environ:
        del os.environ['DUMMY_ENV_VAR']
    try:
        EnvSearchPathEntry('DUMMY_ENV_VAR')
        assert False, 'exception expected'
    except ValueError:
        pass

    assert os.environ['PATH'] == str(EnvSearchPathEntry('PATH'))

    path = SearchPath(
        [SearchPathEntry(p) for p in os.environ['PATH'].split(os.pathsep)])
    assert os.environ['PATH'] == str(path)
    assert os.environ['PATH'].split(os.pathsep) == list(path)

    path2 = SearchPath([EnvSearchPathEntry('PATH')])
    assert os.environ['PATH'].split(os.pathsep) == list(path2)

    path3 = SearchPath([SearchPathEntry('b')])
    path3.insert(0, 'a')
    path3.append('c')
    assert os.pathsep.join('abc') == str(path3)

    path4 = path3 + list('123')
    assert os.pathsep.join('abc123') == str(path4)

    path5 = path3 + SearchPath('123')
    assert os.pathsep.join('abc123') == str(path5)

    assert not SearchPath([])
    assert SearchPath(['a'])



###############################################################################
# (c) Copyright 2018 CERN                                                     #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
__author__ = 'Marco Clemencic <marco.clemencic@cern.ch>'

import sys
import logging


def run(cmd):
    # Prepare the profiler instance
    profiler = None
    try:
        import cProfile
        profiler = cProfile.Profile()
    except ImportError:
        try:
            import profile
            profiler = profile.Profile()
        except ImportError:
            logging.warning(
                'Cannot import cProfile or profile: ignoring --profile')
            exec cmd
    try:
        # if we managed to get the profiler instance, collect profiling stats,
        # and print them on stderr
        profiler.run(cmd)

    except SystemExit:  # intercept normal exit
        # adapted from Python standard profiler.py
        import pstats
        stream = sys.stderr
        stream.write('\n')
        stream.write('{0:=^80}\n\n'.format(' profiling results '))
        stats = pstats.Stats(profiler, stream=stream)
        stats.strip_dirs().sort_stats('cumulative').print_stats(10)
        stream.write((80 * '=') + '\n')

        # propagate return code
        raise



###############################################################################
# (c) Copyright 2018 CERN                                                     #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
'''
@author: Marco Clemencic <marco.clemencic@cern.ch>
'''
import os
from subprocess import check_output, CalledProcessError

ENV_SCRIPTS = [
    '/cvmfs/lhcb.cern.ch/lib/lhcb/LHCBDIRAC/lhcbdirac',
    '/cvmfs/lhcbdev.cern.ch/lib/lhcb/LHCBDIRAC/lhcbdirac'
]


class LHCbDiracEnvError(RuntimeError):
    pass


def listenv(command='true'):
    '''
    Print the environment produced by a command.

    @return list of environment variables as 'name=value'
    '''
    return filter(
        None,
        check_output(
            command + ' >/dev/null && /usr/bin/printenv --null',
            shell=True).split('\x00'))


def envchanges(command):
    '''
    Return the changes produced in the environment by the specified command as
    a set of 'name=value' entries.
    '''
    return set(listenv(command)) - set(listenv())


def envdict(command, changes_only=True):
    '''
    Return a dictionary of environment variable changes (name: value) produced
    by a command.
    '''
    return dict(
        entry.split('=', 1)
        for entry in (envchanges if changes_only else listenv)(command))


def getLHCbDiracEnv(version):
    '''
    Return the changes to the environment required for the specified LHCbDirac
    version, as a dictionary.
    '''
    import logging
    for script in filter(os.path.isfile, ENV_SCRIPTS):
        try:
            logging.getLogger('getLHCbDiracEnv').debug('sourcing "%s %s"',
                                                       script, version)
            return envdict(' '.join(['source', script, version]))
        except CalledProcessError as err:
            logging.debug(str(err))
            pass  # ignore failures of the script (the version may be wrong)
    else:
        raise LHCbDiracEnvError()



###############################################################################
# (c) Copyright 2018 CERN                                                     #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
__author__ = 'Marco Clemencic <marco.clemencic@cern.ch>'

import os
import sys

assert sys.version_info >= (2, 6), "Python >= 2.6 is required"


def _cmakeStr(txt):
    return ' '.join('"%s"' % entry.replace('"', '\\"')
                    for entry in str(txt).split(os.pathsep))


_CMTPROJECTPATH_FMT_SH = ('export CMTPROJECTPATH="{0}'
                          '${{CMTPROJECTPATH:+:${{CMTPROJECTPATH}}}}"\n')
_CMTPROJECTPATH_FMTS = {
    'csh':
    '''if ( $?CMTPROJECTPATH ) then
  setenv CMTPROJECTPATH "{0}:${{CMTPROJECTPATH}}"
else
  setenv CMTPROJECTPATH "{0}"
endif
''',
    'sh':
    _CMTPROJECTPATH_FMT_SH
}


class SearchPathEntry(object):
    def __init__(self, path):
        self.path = path

    def __str__(self):
        return str(self.path)

    def toCMake(self):
        return 'list(INSERT CMAKE_PREFIX_PATH 0 {0})\n'.format(_cmakeStr(self))

    def toCMT(self, shell='sh'):
        # FIXME: should we escape quotes?
        return _CMTPROJECTPATH_FMTS[shell].format(self)

    def __repr__(self):
        return '{0}({1!r})'.format(self.__class__.__name__, self.path)


class EnvSearchPathEntry(SearchPathEntry):
    def __init__(self, envname, default=None):
        if default is None and envname not in os.environ:
            raise ValueError('environment variable %s not defined' % envname)
        self.envname = envname
        self.default = default

    def __str__(self):
        return str(os.environ.get(self.envname, self.default))

    def toCMake(self):
        return (('if(NOT ENV{{{0}}})\n'
                 '  # use a default value\n'
                 '  set(ENV{{{0}}} {1})\n'
                 'endif()\n'
                 'list(INSERT CMAKE_PREFIX_PATH 0 "$ENV{{{0}}}")\n').format(
                     self.envname, _cmakeStr(self)))

    def toCMT(self, shell='sh'):
        if self.envname == 'CMAKE_PREFIX_PATH':
            return ''  # ignore CMAKE_PREFIX_PATH for CMT builds
        if shell == 'csh':
            fmt = ('if ( ! $?{0} ) then\n'
                   '  # use a default value\n'
                   '  setenv {0} "{1}"\n'
                   'endif\n')
        else:
            fmt = ('if [ -z "${0}" ] ; then\n'
                   '  # use a default value\n'
                   '  export {0}="{1}"\n'
                   'fi\n')
        # FIXME: should we escape quotes?
        return (fmt.format(self.envname, self) +
                _CMTPROJECTPATH_FMTS[shell].format('${' + self.envname + '}'))

    def __repr__(self):
        return '{0}({1!r}, {2!r})'.format(self.__class__.__name__,
                                          self.envname, str(self))


def _toEntry(entry):
    '''ensure that the entry is a SearchPathEntry'''
    if isinstance(entry, SearchPathEntry):
        return entry
    return SearchPathEntry(entry)


class SearchPath(object):
    def __init__(self, entries):
        self.entries = list(entries)

    def __iter__(self):
        from itertools import ifilter, chain
        return ifilter(
            None,
            chain.from_iterable(
                str(entry).split(os.pathsep) for entry in self.entries))

    def insert(self, index, value):
        return self.entries.insert(index, value)

    def append(self, value):
        return self.entries.append(value)

    def __setitem__(self, index, value):
        if isinstance(value, SearchPath):
            value = value.entries
        return self.entries.__setitem__(index, value)

    def __getitem__(self, index):
        if isinstance(index, slice):
            return self.__class__(self.entries[index])
        else:
            return self.entries[index]

    def __str__(self):
        return os.pathsep.join(self)

    def __nonzero__(self):
        return bool(self.entries)

    def __add__(self, other):
        if isinstance(other, SearchPath):
            other = other.entries
        return SearchPath(self.entries + other)

    def toCMake(self):
        # we need to add them in reverse order because the 'prepend'
        return '\n'.join(
            _toEntry(entry).toCMake() for entry in self.entries[::-1])

    def toCMT(self, shell='sh'):
        # we need to add them in reverse order because the 'prepend'
        return '\n'.join(
            _toEntry(entry).toCMT(shell) for entry in self.entries[::-1])

    def __repr__(self):
        return '{0}({1!r})'.format(self.__class__.__name__, self.entries)

    def __len__(self):
        return len(self.entries)


def _defaultPath():
    '''
    Return the default search path, based on the variables CMAKE_PREFIX_PATH
    and CMTPROJECTPATH.
    '''
    env_vars = ['CMAKE_PREFIX_PATH', 'CMTPROJECTPATH', 'LHCBPROJECTPATH']
    path = []
    for name in env_vars:
        try:
            path.append(EnvSearchPathEntry(name, ''))
        except ValueError:
            pass  # ignore undefined variables
    return path


path = SearchPath(_defaultPath())


class Error(RuntimeError):
    '''
    Base class for LbRun exceptions.
    '''



#!/usr/bin/env python
from __future__ import print_function
###############################################################################
# (c) Copyright 2018 CERN                                                     #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
'''
Helpers for shell completion functions.
'''
_actions = {}


def action(f):
    '''
    Decorator to flag the list of actions.
    '''
    _actions[f.__name__] = f
    return f


@action
def list_projects():
    'Print the list of known project names'
    from LbEnv import getProjectNames
    print('\n'.join(p for p in sorted(getProjectNames()) if p.upper() != p))


def _get_names(name, flavour='nightly'):
    from json import load
    from urllib2 import urlopen
    from LbEnv.Info import COUCHDB_ROOT
    return load(
        urlopen(
            (COUCHDB_ROOT + '_design/names/_view/{}?group=true').replace(
                'nightlies-nightly', 'nightlies-{}').format(flavour, name),
            timeout=10)).get('rows', [])


@action
def list_nightly_slots():
    'Print the list of nightly slot names used in the last 2 weeks'
    from datetime import date, timedelta
    since = str(date.today() - timedelta(days=14))
    rows = _get_names('slots')
    print('\n'.join(row['key'] for row in rows if row['value'] >= since))


@action
def list_platforms():
    'Print the list of all platforms (from nightlies db)'
    from datetime import date, timedelta
    since = str(date.today() - timedelta(days=14))
    platforms = set(
        row['key'] for row in _get_names('platforms') if row['value'] >= since)
    # FIXME: the list of platform in the release builds includes non released
    # ones
    platforms.update(row['key'] for row in _get_names('platforms', 'release')
                     if '-' in row['key'] and 'cern' not in row['key']
                     and '+o3' not in row['key'])
    print('\n'.join(sorted(platforms)))


@action
def list_compatible_platforms():
    'Print the list of all platforms (from nightlies db)'
    import LbPlatformUtils as lpu
    dirac_platform = lpu.dirac_platform()

    platforms = set(row['key'] for row in _get_names('platforms'))
    platforms.update(row['key'] for row in _get_names('platforms', 'release'))

    print('\n'.join(
        sorted(
            p for p in platforms
            if '-' in p and lpu.can_run(dirac_platform, lpu.requires(p)))))


def main():
    import sys
    from argparse import ArgumentParser
    parser = ArgumentParser()

    parser.add_argument('action', choices=_actions, help='action to call')

    args = parser.parse_args()

    sys.exit(_actions[args.action]())


if __name__ == '__main__':
    main()
